#!/usr/bin/env python

import pprint
import math

import numpy
from osgeo import gdal

import rsgislib
import rsgislib.imageutils
import rsgislib.tools.projection


class StdImgBlockIter:

    """
    A class to read and write images in blocks where all blocks are all identical in
    size. To deal with image boundaries, where block do not divide into the the image
    size, the outside blocks will be overlapped with the previous block being defined
    with reference with the right and bottom edges rather than the left and top edges.
    """

    def __init__(self, img_info_lst, block_size):
        """

        :param img_info_lst:
        :param block_size:

        """
        self.img_info_lst = img_info_lst
        self.n_imgs = len(img_info_lst)
        self.block_size = block_size
        self.img_info = dict()
        self.define_overlaps()
        pprint.pprint(self.img_info)

        self.remain_x_block = False
        self.n_x_blocks = math.floor(self.overlap_width / block_size)
        if (self.n_x_blocks * block_size) < self.overlap_width:
            self.n_x_blocks += 1
            self.remain_x_block = True

        self.remain_y_block = False
        self.n_y_blocks = math.floor(self.overlap_height / block_size)
        if (self.n_y_blocks * block_size) < self.overlap_height:
            self.n_y_blocks += 1
            self.remain_y_block = True

        self.n_blocks = self.n_x_blocks * self.n_y_blocks

        self.c_blocks = 0
        self.c_x_blocks = -1
        self.c_y_blocks = 0

        for img in self.img_info_lst:
            self.img_info[img.name]["n_bands"] = len(img.bands)
            self.img_info[img.name]["np_arr"] = numpy.zeros(
                (len(img.bands), block_size, block_size), dtype=float
            )
            self.img_info[img.name]["dataset"] = gdal.Open(
                img.file_name, gdal.GA_ReadOnly
            )
            if self.img_info[img.name]["dataset"] is None:
                raise rsgislib.RSGISPyException(
                    "Could not open image file: {}".format(img.file_name)
                )
            self.img_info[img.name]["bands"] = dict()
            for band in img.bands:
                self.img_info[img.name]["bands"][band] = self.img_info[img.name][
                    "dataset"
                ].GetRasterBand(band)
                if self.img_info[img.name]["bands"][band] is None:
                    raise rsgislib.RSGISPyException(
                        "Could not open band {} within image {}".format(
                            band, img.file_name
                        )
                    )

        # Output image variables
        self.out_imgs_info = None

    def find_pxl_subset_bbox(self, spat_img_bbox, pxl_res, spat_sub_bbox):
        """

        :param spat_img_bbox:
        :param pxl_res:
        :param spat_sub_bbox:
        :return:

        """
        min_x_diff = spat_sub_bbox[0] - spat_img_bbox[0]
        max_y_diff = spat_img_bbox[3] - spat_sub_bbox[3]

        min_x_pxls = math.floor((min_x_diff / pxl_res[0]) + 0.5)
        max_y_pxls = math.floor((max_y_diff / pxl_res[1]) + 0.5)

        sub_width = math.floor(
            ((spat_sub_bbox[1] - spat_sub_bbox[0]) / pxl_res[0]) + 0.5
        )
        sub_height = math.floor(
            ((spat_sub_bbox[3] - spat_sub_bbox[2]) / pxl_res[1]) + 0.5
        )

        max_x_pxls = min_x_pxls + sub_width
        min_y_pxls = max_y_pxls + sub_height

        return [min_x_pxls, max_x_pxls, min_y_pxls, max_y_pxls]

    def reset_iter(self):
        """
        Reset the variables used for the iterator.

        """
        self.c_blocks = 0
        self.c_x_blocks = -1
        self.c_y_blocks = 0

    def define_overlaps(self):
        """
        An internal function used to calculate the amount of overlap between
        the input images.

        """
        import rsgislib.tools.geometrytools

        bboxes = []
        self.img_epsg = None
        self.img_pxl_res = None
        for img in self.img_info_lst:
            self.img_info[img.name] = dict()
            self.img_info[img.name]["res"] = rsgislib.imageutils.get_img_res(
                img.file_name, abs_vals=True
            )
            self.img_info[img.name]["bbox"] = rsgislib.imageutils.get_img_bbox(
                img.file_name
            )
            self.img_info[img.name]["pxl_size"] = rsgislib.imageutils.get_img_size(
                img.file_name
            )
            self.img_info[img.name][
                "epsg"
            ] = rsgislib.imageutils.get_epsg_proj_from_img(img.file_name)
            if self.img_info[img.name]["epsg"] is None:
                raise rsgislib.RSGISPyException(
                    "The input image ({}) does not have a "
                    "defined projection - please correct.".format(img.file_name)
                )
            if self.img_epsg is None:
                self.img_epsg = self.img_info[img.name]["epsg"]
            else:
                if self.img_epsg != self.img_info[img.name]["epsg"]:
                    raise rsgislib.RSGISPyException(
                        "The input images have different input "
                        "projections: EPSG:{} != EPSG:{}".format(
                            self.img_epsg, self.img_info[img.name]["epsg"]
                        )
                    )

            if self.img_pxl_res is None:
                self.img_pxl_res = self.img_info[img.name]["res"]
            else:
                if (self.img_pxl_res[0] != self.img_info[img.name]["res"][0]) and (
                    self.img_pxl_res[1] != self.img_info[img.name]["res"][1]
                ):
                    raise rsgislib.RSGISPyException(
                        "The input images have different "
                        "input image resolutions: {} != {}".format(
                            self.img_pxl_res, self.img_info[img.name]["res"]
                        )
                    )

            bboxes.append(self.img_info[img.name]["bbox"])

        self.bbox_intersect = rsgislib.tools.geometrytools.bboxes_intersection(bboxes)

        self.overlap_width = 0
        self.overlap_height = 0

        for img in self.img_info_lst:
            self.img_info[img.name]["pxl_bbox"] = self.find_pxl_subset_bbox(
                self.img_info[img.name]["bbox"],
                self.img_info[img.name]["res"],
                self.bbox_intersect,
            )
            if (self.overlap_width == 0) and (self.overlap_height == 0):
                self.overlap_width = (
                    self.img_info[img.name]["pxl_bbox"][1]
                    - self.img_info[img.name]["pxl_bbox"][0]
                )
                self.overlap_height = (
                    self.img_info[img.name]["pxl_bbox"][2]
                    - self.img_info[img.name]["pxl_bbox"][3]
                )
            else:
                overlap_width_tmp = (
                    self.img_info[img.name]["pxl_bbox"][1]
                    - self.img_info[img.name]["pxl_bbox"][0]
                )
                overlap_height_tmp = (
                    self.img_info[img.name]["pxl_bbox"][2]
                    - self.img_info[img.name]["pxl_bbox"][3]
                )
                if self.overlap_width != overlap_width_tmp:
                    raise rsgislib.RSGISPyException("The image widths do not match.")
                if self.overlap_height != overlap_height_tmp:
                    raise rsgislib.RSGISPyException("The image height do not match.")

    def __iter__(self):
        """
        Returns an instances of itself.
        """
        return self

    def __next__(self):
        """
        The key function for the iterator used to read the image blocks.

        :return: x_block, y_block, img_data (dict with image name as key; each image
                 has a np_arr variable for the image data.

        """
        if self.c_blocks < self.n_blocks:
            if self.c_x_blocks < (self.n_x_blocks - 1):
                self.c_x_blocks += 1
            else:
                self.c_x_blocks = 0
                self.c_y_blocks += 1
            self.c_blocks += 1

            out_img_data = dict()

            if self.c_y_blocks == (self.n_y_blocks - 1):
                if self.c_x_blocks == (self.n_x_blocks - 1):
                    # X,Y End Case
                    for img in self.img_info_lst:
                        out_img_data[img.name] = dict()
                        out_img_data[img.name]["pxl_x"] = (
                            self.img_info[img.name]["pxl_bbox"][1] - self.block_size
                        )
                        out_img_data[img.name]["pxl_y"] = (
                            self.img_info[img.name]["pxl_bbox"][2] - self.block_size
                        )
                else:
                    # Y End Case
                    for img in self.img_info_lst:
                        out_img_data[img.name] = dict()
                        out_img_data[img.name]["pxl_x"] = self.img_info[img.name][
                            "pxl_bbox"
                        ][0] + (self.block_size * self.c_x_blocks)
                        out_img_data[img.name]["pxl_y"] = (
                            self.img_info[img.name]["pxl_bbox"][2] - self.block_size
                        )
            elif self.c_x_blocks == (self.n_x_blocks - 1):
                # X End Case.
                for img in self.img_info_lst:
                    out_img_data[img.name] = dict()
                    out_img_data[img.name]["pxl_x"] = (
                        self.img_info[img.name]["pxl_bbox"][1] - self.block_size
                    )
                    out_img_data[img.name]["pxl_y"] = self.img_info[img.name][
                        "pxl_bbox"
                    ][3] + (self.block_size * self.c_y_blocks)
            else:
                # Not End Case
                for img in self.img_info_lst:
                    out_img_data[img.name] = dict()
                    out_img_data[img.name]["pxl_x"] = self.img_info[img.name][
                        "pxl_bbox"
                    ][0] + (self.block_size * self.c_x_blocks)
                    out_img_data[img.name]["pxl_y"] = self.img_info[img.name][
                        "pxl_bbox"
                    ][3] + (self.block_size * self.c_y_blocks)

            for img in self.img_info_lst:
                for band in img.bands:
                    self.img_info[img.name]["bands"][band].ReadAsArray(
                        xoff=out_img_data[img.name]["pxl_x"],
                        yoff=out_img_data[img.name]["pxl_y"],
                        win_xsize=self.block_size,
                        win_ysize=self.block_size,
                        buf_obj=self.img_info[img.name]["np_arr"][band - 1],
                    )
                out_img_data[img.name]["np_arr"] = self.img_info[img.name]["np_arr"]

            return (self.c_x_blocks, self.c_y_blocks, out_img_data)
        raise StopIteration

    def __len__(self):
        """
        A function which returns the length of the iterator (i.e., number of blocks).

        :return: number of blocks.

        """
        return self.n_blocks

    def create_output_imgs(self, out_imgs_info):
        """
        A function to create the output image(s) if required during processing.

        :param out_imgs_info: a list of rsgislib.imageutils.OutImageInfo
                              objects specifying the output image(s).

        """
        out_wkt_str = rsgislib.tools.projection.get_wkt_from_epsg_code(
            int(self.img_epsg)
        )
        if out_wkt_str is None:
            raise rsgislib.RSGISPyException(
                "Did not have a projection string for the EPSG code: {}".format(
                    self.img_epsg
                )
            )
        if math.fabs(self.img_pxl_res[0]) != math.fabs(self.img_pxl_res[1]):
            raise rsgislib.RSGISPyException(
                "Currently limited to only outputting images with square pixels."
            )
        out_img_res = math.fabs(self.img_pxl_res[0])
        self.out_imgs_info = out_imgs_info
        self.out_imgs_objs = dict()
        for img in out_imgs_info:
            rsgislib.imageutils.create_blank_img_from_bbox(
                self.bbox_intersect,
                out_wkt_str,
                img.file_name,
                out_img_res,
                img.no_data_val,
                img.nbands,
                img.gdalformat,
                img.datatype,
                snap_to_grid=False,
            )

            self.out_imgs_objs[img.name] = dict()
            self.out_imgs_objs[img.name]["dataset"] = gdal.Open(
                img.file_name, gdal.GA_Update
            )
            if self.out_imgs_objs[img.name]["dataset"] is None:
                raise rsgislib.RSGISPyException(
                    "Could not open image file: {}".format(img.file_name)
                )
            self.out_imgs_objs[img.name]["nbands"] = img.nbands
            self.out_imgs_objs[img.name]["bands"] = dict()
            for band in range(img.nbands):
                self.out_imgs_objs[img.name]["bands"][band + 1] = self.out_imgs_objs[
                    img.name
                ]["dataset"].GetRasterBand(band + 1)
                if self.out_imgs_objs[img.name]["bands"][band + 1] is None:
                    raise rsgislib.RSGISPyException(
                        "Could not open band {} within image {}".format(
                            band + 1, img.file_name
                        )
                    )

    def write_block_to_image(self, x_block, y_block, img_blocks):
        """
        A function to write the processed blocks to an output image.

        :param x_block: The X index of the block being written.
        :param y_block: The Y index of the block being written.
        :param img_blocks: A dict of output image block data. Keys must match
                           the name of the images. Note. data block data returned
                           must have shape (nbands, xsize, ysize)

        """
        if y_block == (self.n_y_blocks - 1):
            if x_block == (self.n_x_blocks - 1):
                # X,Y End Case
                pxl_x = self.overlap_width - self.block_size
                pxl_y = self.overlap_height - self.block_size
            else:
                # Y End Case
                pxl_x = self.block_size * x_block
                pxl_y = self.overlap_height - self.block_size
        elif x_block == (self.n_x_blocks - 1):
            # X End Case.
            pxl_x = self.overlap_width - self.block_size
            pxl_y = self.block_size * y_block
        else:
            # Not End Case
            pxl_x = self.block_size * x_block
            pxl_y = self.block_size * y_block

        for img in img_blocks:
            if img in self.out_imgs_objs:
                block_shp = img_blocks[img].shape
                if block_shp[0] != self.out_imgs_objs[img]["nbands"]:
                    raise rsgislib.RSGISPyException(
                        "The number of image bands in the output file "
                        "and returned data block do not match "
                        "(block:{}; image:{})".format(
                            block_shp[0], self.out_imgs_objs[img]["nbands"]
                        )
                    )
                if (block_shp[1] != self.block_size) and (
                    block_shp[2] != self.block_size
                ):
                    raise rsgislib.RSGISPyException(
                        "The block size is either not square or the "
                        "same size as the size expected (block: {} x {}; "
                        "parameterised block: {}).".format(
                            block_shp[1], block_shp[2], self.block_size
                        )
                    )
                for band in range(block_shp[0]):
                    self.out_imgs_objs[img]["bands"][band + 1].WriteArray(
                        img_blocks[img][band], pxl_x, pxl_y
                    )


"""
import tqdm

imgBandInfo = []
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='S2_UVD_27sept_27700_sub.kea', name='sen2a', bands=[1,2,3,4,5,6,7,8,9,10]))
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='S2_UVD_27sept_27700_sub_roi1.kea', name='sen2a1', bands=[1,2,3,4,5,6,7,8,9,10]))
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='S2_UVD_27sept_27700_sub_roi2.kea', name='sen2a2', bands=[1,2,3,4,5,6,7,8,9,10]))

imgOutInfo = []
imgOutInfo.append(OutImageInfo(file_name='S2_UVD_27sept_27700_sub_ndvi.kea', name='sen2a_ndvi', nbands=1, no_data_val=-999, gdalformat='KEA', datatype=rsgislib.TYPE_32FLOAT))
imgOutInfo.append(OutImageInfo(file_name='S2_UVD_27sept_27700_sub_ndwi.kea', name='sen2a_ndwi', nbands=1, no_data_val=-999, gdalformat='KEA', datatype=rsgislib.TYPE_32FLOAT))

img_iter = StdImgBlockIter(imgBandInfo, 64)
img_iter.create_output_imgs(imgOutInfo)
for x_block, y_block, out_img_data in tqdm.tqdm(img_iter):
    hello = 'world'
    #print("Tile: ({} x {})".format(x_block, y_block))
    #for img in out_img_data:
    #    print("Mean = {}".format(numpy.mean(out_img_data[img]['np_arr'])))
    ndvi = (out_img_data['sen2a']['np_arr'][8] - out_img_data['sen2a']['np_arr'][3]) / (out_img_data['sen2a']['np_arr'][8] + out_img_data['sen2a']['np_arr'][3])
    ndwi = (out_img_data['sen2a']['np_arr'][8] - out_img_data['sen2a']['np_arr'][9]) / (out_img_data['sen2a']['np_arr'][8] + out_img_data['sen2a']['np_arr'][9])
    out_data = {'sen2a_ndvi': numpy.expand_dims(ndvi, axis=0), 'sen2a_ndwi': numpy.expand_dims(ndwi, axis=0)}
    #print(ndvi.shape)
    img_iter.write_block_to_image(x_block, y_block, out_data)

"""



#! /usr/bin/env python
############################################################################
#  tmask.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provides functions for applying the tmask cloud masking algorithm.
#           See https://github.com/klh5/rios_tmask
#
# Zhu, Z. and Woodcock, C.E. Automated cloud, cloud shadow, and snow detection
# in multitemporal Landsat data: An algorithm designed specifically for monitoring
# land cover change. Remote Sensing of Environment. 2014, 152, 217–234.
# doi:10.1016/j.rse.2014.06.012.
#
# Author: Katie Awty-Carroll (ed by Pete Bunting)
# Email: petebunting@mac.com
# Date: 24/2/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import json
import sys
from datetime import datetime

import numpy
import statsmodels.api

from rios import fileinfo
from rios import applier


class RLMRemoveOutliers(object):
    def __init__(self, toa_data, results):

        self.T = 365.25
        self.pi_val = (2 * numpy.pi) / self.T
        self.toa_data = toa_data
        self.results = results

        self.pi_val_change = None
        self.green_model = None
        self.nir_model = None
        self.swir_model = None
        self.b2_delta = None
        self.b4_delta = None
        self.b5_delta = None

    def cleanData(self, num_years, threshold):
        """
        Uses the green, NIR, and SWIR band data to build RLMs for detecting cloud and snow outliers

        :param num_years:
        :param threshold:
        :return:
        """
        self.pi_val_change = (2 * numpy.pi) / (num_years * self.T)
        start_date = self.toa_data[:, 0].min()
        rescaled = self.toa_data[:, 0] - start_date

        try:
            # Get coefficients for the three models
            self.green_model, self.b2_delta = self.makeRLMModel(
                rescaled, self.toa_data[:, 1]
            )
            self.nir_model, self.b4_delta = self.makeRLMModel(
                rescaled, self.toa_data[:, 2]
            )
            self.swir_model, self.b5_delta = self.makeRLMModel(
                rescaled, self.toa_data[:, 3]
            )

            self.toa_data = numpy.hstack((self.toa_data, self.b2_delta.reshape(-1, 1)))
            self.toa_data = numpy.hstack((self.toa_data, self.b4_delta.reshape(-1, 1)))
            self.toa_data = numpy.hstack((self.toa_data, self.b5_delta.reshape(-1, 1)))

            self.dropOutliers(threshold)
        except numpy.linalg.LinAlgError:
            pass
        return self.results

    def makeRLMModel(self, datetimes, band_data):
        """
        Builds the model and stores the coefficients

        :param datetimes:
        :param band_data:
        :return:

        """
        x = numpy.array(
            [
                numpy.ones_like(datetimes),  # Add constant
                numpy.cos(self.pi_val * datetimes),
                numpy.sin(self.pi_val * datetimes),
                numpy.cos(self.pi_val_change * datetimes),
                numpy.sin(self.pi_val_change * datetimes),
            ]
        ).T

        rlm_model = statsmodels.api.RLM(
            band_data, x, M=statsmodels.api.robust.norms.TukeyBiweight(c=0.4685)
        )
        rlm_result = rlm_model.fit(maxiter=5)
        delta = band_data - rlm_result.predict(x)
        return rlm_result, delta

    def dropOutliers(self, threshold):
        """

        :param threshold:

        """
        self.results = numpy.where(
            (
                (self.toa_data[:, 4] < threshold)
                & (
                    (self.toa_data[:, 5] > -threshold)
                    | (self.toa_data[:, 6] > -threshold)
                )
            ),
            self.results,
            self.results + 1,
        )


def _gen_tmask(info, inputs, outputs, other_args):
    """
    Run per-block by RIOS. In this case each block is a
    single pixel. Given a block of values for each band for each date, returns
    a numpy array containing a mask where screened out data = 1 and data
    to keep = 0.

    """
    dates = other_args.dates
    nodata = other_args.nodata

    # Set up default output
    # Assumes all pixels are clear
    results = numpy.zeros(len(inputs.images), dtype="uint8")

    green_vals = numpy.array(
        [[inputs.images[t][0][0][0]] for t in range(0, len(inputs.images))]
    )
    nir_vals = numpy.array(
        [[inputs.images[t][1][0][0]] for t in range(0, len(inputs.images))]
    )
    swir1_vals = numpy.array(
        [[inputs.images[t][2][0][0]] for t in range(0, len(inputs.images))]
    )

    model_inputs = numpy.hstack((other_args.dates, green_vals, nir_vals, swir1_vals))
    drop_indices = numpy.where(numpy.any(model_inputs == nodata, axis=1))
    drop_indices = numpy.array(drop_indices).reshape(-1)

    # Remove any rows where all band values are 0
    model_inputs = model_inputs[numpy.all(model_inputs != nodata, axis=1)]

    # Need a minimum of 12 observations
    if len(model_inputs) >= 12:
        # Output array needs to be matched in size to input array
        output_arr = numpy.delete(results, drop_indices)
        tmask = RLMRemoveOutliers(model_inputs, output_arr)
        num_years = numpy.ceil((numpy.max(dates) - numpy.min(dates)) / 365)
        output_arr = tmask.cleanData(num_years, other_args.threshold)
        results = numpy.insert(
            output_arr, numpy.array(drop_indices), numpy.zeros(len(drop_indices))
        )

    results = results.reshape(len(results), 1, 1, 1)
    outputs.outimage = results


def run_tmask(
    json_fp,
    gdal_format="KEA",
    num_processes=1,
    green_band=2,
    nir_band=4,
    swir_band=5,
    threshold=40,
):
    """
    Main function to run to generate the output masks. Given an input JSON file,
    generates a mask for each date where 1=cloud/cloud shadow/snow and 0=clear.
    Opening/closing of files, generation of blocks and use of multiprocessing is
    all handled by RIOS.

    A minimum of 12 observations is required to create the masks.

    :param json_fp: Path to JSON file which provides a dictionary where for each
                    date, an input file name and an output file name are provided.
    :param gdal_format: The file format of the output image (e.g., KEA, GTIFF). (Default: KEA)
    :param num_processes: Number of concurrent processes to use. (Default: 1)
    :param green_band: GDAL band number for green spectral band. Defaults to 2.
    :param nir_band: GDAL band number for NIR spectral band. Defaults to 4.
    :param swir_band: GDAL band number for SWIR spectral band. Defaults to 5.
    :param threshold: Numerical threshold for screening out cloud, cloud shadow, and snow.
                      Defaults to 40. See Zhu & Woodcock (2014) for details.

    """
    ip_paths = []
    op_paths = []
    dates = []

    try:
        # Open and read JSON file containing date:filepath pairs
        with open(json_fp) as json_file:
            image_list = json.load(json_file)

            for date in image_list.items():
                dates.append([datetime.strptime(date[0], "%Y-%m-%d").toordinal()])
                ip_paths.append(date[1]["input"])
                op_paths.append(date[1]["output"])
    except FileNotFoundError:
        print("Could not find the provided JSON file.")
        sys.exit()
    except json.decoder.JSONDecodeError as e:
        print("There is an error in the provided JSON file: {}".format(e))
        sys.exit()

    # Create object to hold input files
    infiles = applier.FilenameAssociations()
    infiles.images = ip_paths

    # Create object to hold output file
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = op_paths

    # ApplierControls object holds details on how processing should be done
    app = applier.ApplierControls()

    # Set window size to 1 because we are working per-pixel
    app.setWindowXsize(1)
    app.setWindowYsize(1)

    # Set output file type
    app.setOutputDriverName(gdal_format)

    # Use Python's multiprocessing module
    app.setJobManagerType("multiprocessing")
    app.setNumThreads(num_processes)

    # Open first image in list to use as a template
    template_image = fileinfo.ImageInfo(infiles.images[0])

    # Get no data value
    nodata = template_image.nodataval[0]

    # Need to tell the applier to only use the specified bands
    app.selectInputImageLayers([green_band, nir_band, swir_band])

    # Set up output layer name
    app.setLayerNames(["tmask"])

    # Additional arguments - have to be passed as a single object
    other_args = applier.OtherInputs()
    other_args.dates = dates
    other_args.threshold = threshold
    other_args.nodata = nodata

    try:
        applier.apply(_gen_tmask, infiles, outfiles, otherArgs=other_args, controls=app)
    except RuntimeError as e:
        print("There was an error processing the images: {}".format(e))
        print("Do all images in the JSON file exist?")



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
IMGUTILS_DATA_DIR = os.path.join(DATA_DIR, "imageutils")


def test_create_img_extent_lut(tmp_path):
    import rsgislib.imageutils.imagelut
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    assert os.path.exists(vec_file)


def test_create_img_extent_lut_wgs84(tmp_path):
    import rsgislib.imageutils.imagelut
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=True,
        overwrite_lut_file=False,
    )

    assert os.path.exists(vec_file)


def test_get_all_lut_imgs(tmp_path):
    import rsgislib.imageutils.imagelut
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    rsgislib.imageutils.imagelut.get_all_lut_imgs(vec_file, vec_lyr)


def test_query_img_lut(tmp_path):
    import rsgislib.imageutils.imagelut
    import rsgislib.imageutils
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    scn_bbox = rsgislib.imageutils.get_img_bbox(
        os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    )

    rsgislib.imageutils.imagelut.query_img_lut(scn_bbox, vec_file, vec_lyr)


def test_get_raster_lyr(tmp_path):
    import rsgislib.imageutils.imagelut
    import rsgislib.imageutils
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    scn_bbox = rsgislib.imageutils.get_img_bbox(
        os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    )

    out_img = rsgislib.imageutils.imagelut.get_raster_lyr(
        scn_bbox, vec_file, vec_lyr, tmp_path
    )
    assert os.path.exists(out_img)


def test_query_file_lut(tmp_path):
    import rsgislib.imageutils.imagelut
    import rsgislib.imageutils
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    roi_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_lyr = "aber_osgb_single_poly"

    rsgislib.imageutils.imagelut.query_file_lut(
        vec_file, vec_lyr, roi_file, roi_lyr, None, False, False
    )


def test_query_file_lut_gztar(tmp_path):
    import rsgislib.imageutils.imagelut
    import rsgislib.imageutils
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    roi_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_lyr = "aber_osgb_single_poly"

    out_file = os.path.join(tmp_path, "out.tar.gz")
    rsgislib.imageutils.imagelut.query_file_lut(
        vec_file, vec_lyr, roi_file, roi_lyr, out_file, True, False
    )


def test_query_file_lut_cp(tmp_path):
    import rsgislib.imageutils.imagelut
    import rsgislib.imageutils
    import glob

    input_imgs = glob.glob(os.path.join(IMGUTILS_DATA_DIR, "s2_tiles", "*.kea"))
    vec_file = os.path.join(tmp_path, "test_lut.gpkg")
    vec_lyr = "test_lut"
    rsgislib.imageutils.imagelut.create_img_extent_lut(
        input_imgs,
        vec_file,
        vec_lyr,
        "GPKG",
        ignore_none_imgs=False,
        out_proj_wgs84=False,
        overwrite_lut_file=False,
    )

    roi_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    roi_lyr = "aber_osgb_single_poly"

    rsgislib.imageutils.imagelut.query_file_lut(
        vec_file, vec_lyr, roi_file, roi_lyr, tmp_path, False, True
    )



#! /usr/bin/env python
############################################################################
#  classlightgbm.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 05/11/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import rsgislib
import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis

import numpy

import h5py

from osgeo import gdal

from rios import applier
from rios import cuiprogress
from rios import rat

HAVE_LIGHTGBM = True
try:
    import lightgbm as lgb
except ImportError:
    HAVE_LIGHTGBM = False

from sklearn.metrics import roc_auc_score
from sklearn.metrics import accuracy_score

import gc

import json


def optimise_lightgbm_binary_classifier(
    out_params_file,
    cls1_train_file,
    cls1_valid_file,
    cls2_train_file,
    cls2_valid_file,
    unbalanced=False,
    n_threads=1,
    scale_pos_weight=None,
    early_stopping_rounds=100,
    num_iterations=5000,
    num_boost_round=100,
    max_n_leaves=50,
    learning_rate=0.05,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary lightgbm
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that lightgbm and skopt modules to be installed.

    :param out_params_file: The output model parameters which have been optimised.
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param unbalanced: Specify that the training data is unbalance (i.e., a different number of samples per class)
                       and LightGBM will try to take this into account during training.
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")
    from skopt.space import Real, Integer
    from skopt import gp_minimize

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Finished Reading Data")

    d_train = lgb.Dataset(
        [train_cls2, train_cls1],
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )
    d_valid = lgb.Dataset(
        [valid_cls2, valid_cls1],
        label=numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl)),
    )

    vaild_np = numpy.concatenate((valid_cls2, valid_cls1))
    vaild_lbl_np = numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl))

    space = [
        Integer(3, 10, name="max_depth"),
        Integer(6, max_n_leaves, name="num_leaves"),
        Integer(3, 50, name="min_data_in_leaf"),
        Real(0, 5, name="lambda_l1"),
        Real(0, 3, name="lambda_l2"),
        Real(0.1, 0.9, name="feature_fraction"),
        Real(0.8, 1.0, name="bagging_fraction"),
        Real(0.001, 0.1, name="min_split_gain"),
        Real(1, 50, name="min_child_weight"),
        Real(1, 1.2, name="reg_alpha"),
        Real(1, 1.4, name="reg_lambda"),
    ]

    if scale_pos_weight is None:
        scale_pos_weight = num_cls2_train_rows / num_cls1_train_rows
        if scale_pos_weight < 1:
            scale_pos_weight = 1
    print("scale_pos_weight = {}".format(scale_pos_weight))

    def _objective(values):
        if unbalanced:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "auc,binary_error",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "binary",
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "boost_from_average": True,
                "is_unbalance": True,
                "verbose": -1,
            }
        else:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "auc,binary_error",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "binary",
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "scale_pos_weight": scale_pos_weight,
                "boost_from_average": True,
                "is_unbalance": False,
                "verbose": -1,
            }

        print("\nNext set of params.....", params)

        evals_results = {}
        model_lgb = lgb.train(
            params,
            d_train,
            valid_sets=[d_train, d_valid],
            valid_names=["train", "valid"],
            evals_result=evals_results,
            num_boost_round=num_boost_round,
            early_stopping_rounds=early_stopping_rounds,
            verbose_eval=None,
            feval=None,
            init_model=mdl_cls_obj,
        )

        auc = -roc_auc_score(vaild_lbl_np, model_lgb.predict(vaild_np))
        print("\nAUROC.....", -auc, ".....iter.....", model_lgb.current_iteration())
        gc.collect()
        return auc

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))

    best_params = res_gp.x

    params = {
        "max_depth": int(best_params[0]),
        "num_leaves": int(best_params[1]),
        "min_data_in_leaf": int(best_params[2]),
        "feature_pre_filter": False,
        "lambda_l1": float(best_params[3]),
        "lambda_l2": float(best_params[4]),
        "metric": "auc,binary_error",
        "boosting_type": "gbdt",
        "objective": "binary",
        "feature_fraction": float(best_params[5]),
        "bagging_fraction": float(best_params[6]),
        "min_split_gain": float(best_params[7]),
        "min_child_weight": float(best_params[8]),
        "reg_alpha": float(best_params[9]),
        "reg_lambda": float(best_params[10]),
        "scale_pos_weight": float(scale_pos_weight),
        "boost_from_average": True,
        "is_unbalance": unbalanced,
    }

    with open(out_params_file, "w") as fp:
        json.dump(
            params,
            fp,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )


def train_lightgbm_binary_classifier(
    out_mdl_file,
    cls_params_file,
    cls1_train_file,
    cls1_valid_file,
    cls1_test_file,
    cls2_train_file,
    cls2_valid_file,
    cls2_test_file,
    unbalanced=False,
    n_threads=1,
    scale_pos_weight=None,
    early_stopping_rounds=100,
    num_iterations=5000,
    num_boost_round=100,
    learning_rate=0.05,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary lightgbm
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that lightgbm and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_params_file: A JSON file with the model parameters
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param unbalanced: Specify that the training data is unbalance (i.e., a different number of samples per class)
                       and LightGBM will try to take this into account during training.
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 1 Testing")
    f = h5py.File(cls1_test_file, "r")
    num_cls1_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_test_rows = {}".format(num_cls1_test_rows))
    test_cls1 = numpy.array(f["DATA/DATA"])
    test_cls1_lbl = numpy.ones(num_cls1_test_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Reading Class 2 Testing")
    f = h5py.File(cls2_test_file, "r")
    num_cls2_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_test_rows = {}".format(num_cls2_test_rows))
    test_cls2 = numpy.array(f["DATA/DATA"])
    test_cls2_lbl = numpy.zeros(num_cls2_test_rows, dtype=int)

    print("Finished Reading Data")

    d_train = lgb.Dataset(
        [train_cls2, train_cls1],
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )
    d_valid = lgb.Dataset(
        [valid_cls2, valid_cls1],
        label=numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl)),
    )

    test_np = numpy.concatenate((test_cls2, test_cls1))
    test_lbl_np = numpy.concatenate((test_cls2_lbl, test_cls1_lbl))

    if scale_pos_weight is None:
        scale_pos_weight = num_cls2_train_rows / num_cls1_train_rows
        if scale_pos_weight < 1:
            scale_pos_weight = 1
    print("scale_pos_weight = {}".format(scale_pos_weight))

    with open(cls_params_file) as f:
        cls_params = json.load(f)

    print("Start Training Find Classifier")

    if unbalanced:
        params = {
            "max_depth": cls_params["max_depth"],
            "num_leaves": cls_params["num_leaves"],
            "min_data_in_leaf": cls_params["min_data_in_leaf"],
            "feature_pre_filter": False,
            "lambda_l1": cls_params["lambda_l1"],
            "lambda_l2": cls_params["lambda_l2"],
            "metric": "auc,binary_error",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "binary",
            "learning_rate": learning_rate,
            "feature_fraction": cls_params["feature_fraction"],
            "bagging_fraction": cls_params["bagging_fraction"],
            "min_split_gain": cls_params["min_split_gain"],
            "min_child_weight": cls_params["min_child_weight"],
            "reg_alpha": cls_params["reg_alpha"],
            "reg_lambda": cls_params["reg_lambda"],
            "num_iterations": num_iterations,
            "scale_pos_weight": scale_pos_weight,
            "boost_from_average": True,
            "is_unbalance": True,
            "verbose": -1,
        }
    else:
        params = {
            "max_depth": cls_params["max_depth"],
            "num_leaves": cls_params["num_leaves"],
            "min_data_in_leaf": cls_params["min_data_in_leaf"],
            "feature_pre_filter": False,
            "lambda_l1": cls_params["lambda_l1"],
            "lambda_l2": cls_params["lambda_l2"],
            "metric": "auc,binary_error",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "binary",
            "learning_rate": learning_rate,
            "feature_fraction": cls_params["feature_fraction"],
            "bagging_fraction": cls_params["bagging_fraction"],
            "min_split_gain": cls_params["min_split_gain"],
            "min_child_weight": cls_params["min_child_weight"],
            "reg_alpha": cls_params["reg_alpha"],
            "reg_lambda": cls_params["reg_lambda"],
            "num_iterations": num_iterations,
            "scale_pos_weight": scale_pos_weight,
            "boost_from_average": True,
            "is_unbalance": False,
            "verbose": -1,
        }

    evals_results = {}
    model = lgb.train(
        params,
        d_train,
        valid_sets=[d_train, d_valid],
        valid_names=["train", "valid"],
        evals_result=evals_results,
        num_boost_round=num_boost_round,
        early_stopping_rounds=early_stopping_rounds,
        verbose_eval=None,
        feval=None,
        init_model=mdl_cls_obj,
    )
    test_auc = roc_auc_score(test_lbl_np, model.predict(test_np))
    print("Testing AUC: {}".format(test_auc))
    print("Finish Training")

    model.save_model(out_mdl_file)

    pred_test = model.predict(test_np)
    for i in range(test_np.shape[0]):
        if pred_test[i] >= 0.5:
            pred_test[i] = 1
        else:
            pred_test[i] = 0
    len(pred_test)

    test_acc = accuracy_score(test_lbl_np, pred_test)
    print("Testing Accuracy: {}".format(test_acc))


def train_opt_lightgbm_binary_classifier(
    out_mdl_file,
    out_params_file,
    cls1_train_file,
    cls1_valid_file,
    cls1_test_file,
    cls2_train_file,
    cls2_valid_file,
    cls2_test_file,
    unbalanced=False,
    n_threads=1,
    scale_pos_weight=None,
    early_stopping_rounds=100,
    num_iterations=5000,
    num_boost_round=100,
    learning_rate=0.05,
    max_n_leaves=50,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a binary lightgbm
    classifier. Class 1 is the class which you are interested in and Class 2 is the 'other class'.

    This function requires that lightgbm and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param out_params_file: The output model parameters which have been optimised.
    :param cls1_train_file: Training samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_valid_file: Validation samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls1_test_file: Testing samples HDF5 file for the primary class (i.e., the one being classified)
    :param cls2_train_file: Training samples HDF5 file for the 'other' class
    :param cls2_valid_file: Validation samples HDF5 file for the 'other' class
    :param cls2_test_file: Testing samples HDF5 file for the 'other' class
    :param unbalanced: Specify that the training data is unbalance (i.e., a different number of samples per class)
                       and LightGBM will try to take this into account during training.
    :param n_threads: The number of threads to use for the training.
    :param scale_pos_weight: Optional, default is None. If None then a value will automatically be calculated.
                             Parameter used to balance imbalanced training data.

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    print("Reading Class 1 Training")
    f = h5py.File(cls1_train_file, "r")
    num_cls1_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_train_rows = {}".format(num_cls1_train_rows))
    train_cls1 = numpy.array(f["DATA/DATA"])
    train_cls1_lbl = numpy.ones(num_cls1_train_rows, dtype=int)

    print("Reading Class 1 Validation")
    f = h5py.File(cls1_valid_file, "r")
    num_cls1_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_valid_rows = {}".format(num_cls1_valid_rows))
    valid_cls1 = numpy.array(f["DATA/DATA"])
    valid_cls1_lbl = numpy.ones(num_cls1_valid_rows, dtype=int)

    print("Reading Class 1 Testing")
    f = h5py.File(cls1_test_file, "r")
    num_cls1_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls1_test_rows = {}".format(num_cls1_test_rows))
    test_cls1 = numpy.array(f["DATA/DATA"])
    test_cls1_lbl = numpy.ones(num_cls1_test_rows, dtype=int)

    print("Reading Class 2 Training")
    f = h5py.File(cls2_train_file, "r")
    num_cls2_train_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_train_rows = {}".format(num_cls2_train_rows))
    train_cls2 = numpy.array(f["DATA/DATA"])
    train_cls2_lbl = numpy.zeros(num_cls2_train_rows, dtype=int)

    print("Reading Class 2 Validation")
    f = h5py.File(cls2_valid_file, "r")
    num_cls2_valid_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_valid_rows = {}".format(num_cls2_valid_rows))
    valid_cls2 = numpy.array(f["DATA/DATA"])
    valid_cls2_lbl = numpy.zeros(num_cls2_valid_rows, dtype=int)

    print("Reading Class 2 Testing")
    f = h5py.File(cls2_test_file, "r")
    num_cls2_test_rows = f["DATA/DATA"].shape[0]
    print("num_cls2_test_rows = {}".format(num_cls2_test_rows))
    test_cls2 = numpy.array(f["DATA/DATA"])
    test_cls2_lbl = numpy.zeros(num_cls2_test_rows, dtype=int)

    print("Finished Reading Data")

    d_train = lgb.Dataset(
        [train_cls2, train_cls1],
        label=numpy.concatenate((train_cls2_lbl, train_cls1_lbl)),
    )
    d_valid = lgb.Dataset(
        [valid_cls2, valid_cls1],
        label=numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl)),
    )

    vaild_np = numpy.concatenate((valid_cls2, valid_cls1))
    vaild_lbl_np = numpy.concatenate((valid_cls2_lbl, valid_cls1_lbl))

    test_np = numpy.concatenate((test_cls2, test_cls1))
    test_lbl_np = numpy.concatenate((test_cls2_lbl, test_cls1_lbl))

    space = [
        Integer(3, 10, name="max_depth"),
        Integer(6, max_n_leaves, name="num_leaves"),
        Integer(3, 50, name="min_data_in_leaf"),
        Real(0, 5, name="lambda_l1"),
        Real(0, 3, name="lambda_l2"),
        Real(0.1, 0.9, name="feature_fraction"),
        Real(0.8, 1.0, name="bagging_fraction"),
        Real(0.001, 0.1, name="min_split_gain"),
        Real(1, 50, name="min_child_weight"),
        Real(1, 1.2, name="reg_alpha"),
        Real(1, 1.4, name="reg_lambda"),
    ]

    if scale_pos_weight is None:
        scale_pos_weight = num_cls2_train_rows / num_cls1_train_rows
        if scale_pos_weight < 1:
            scale_pos_weight = 1
    print("scale_pos_weight = {}".format(scale_pos_weight))

    def _objective(values):
        if unbalanced:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "auc,binary_error",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "binary",
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "boost_from_average": True,
                "is_unbalance": True,
                "verbose": -1,
            }
        else:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "auc,binary_error",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "binary",
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "scale_pos_weight": scale_pos_weight,
                "boost_from_average": True,
                "is_unbalance": False,
                "verbose": -1,
            }

        print("\nNext set of params.....", params)

        evals_results = {}
        model_lgb = lgb.train(
            params,
            d_train,
            valid_sets=[d_train, d_valid],
            valid_names=["train", "valid"],
            evals_result=evals_results,
            num_boost_round=num_boost_round,
            early_stopping_rounds=early_stopping_rounds,
            verbose_eval=None,
            feval=None,
            init_model=mdl_cls_obj,
        )

        auc = -roc_auc_score(vaild_lbl_np, model_lgb.predict(vaild_np))
        print("\nAUROC.....", -auc, ".....iter.....", model_lgb.current_iteration())
        gc.collect()
        return auc

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))

    best_params = res_gp.x

    params = {
        "max_depth": int(best_params[0]),
        "num_leaves": int(best_params[1]),
        "min_data_in_leaf": int(best_params[2]),
        "feature_pre_filter": False,
        "lambda_l1": float(best_params[3]),
        "lambda_l2": float(best_params[4]),
        "metric": "auc,binary_error",
        "boosting_type": "gbdt",
        "objective": "binary",
        "feature_fraction": float(best_params[5]),
        "bagging_fraction": float(best_params[6]),
        "min_split_gain": float(best_params[7]),
        "min_child_weight": float(best_params[8]),
        "reg_alpha": float(best_params[9]),
        "reg_lambda": float(best_params[10]),
        "scale_pos_weight": float(scale_pos_weight),
        "boost_from_average": True,
        "is_unbalance": unbalanced,
    }

    with open(out_params_file, "w") as fp:
        json.dump(
            params,
            fp,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )

    print("Start Training Find Classifier")

    if unbalanced:
        params = {
            "max_depth": best_params[0],
            "num_leaves": best_params[1],
            "min_data_in_leaf": best_params[2],
            "feature_pre_filter": False,
            "lambda_l1": best_params[3],
            "lambda_l2": best_params[4],
            "metric": "auc,binary_error",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "binary",
            "learning_rate": learning_rate,
            "feature_fraction": best_params[5],
            "bagging_fraction": best_params[6],
            "min_split_gain": best_params[7],
            "min_child_weight": best_params[8],
            "reg_alpha": best_params[9],
            "reg_lambda": best_params[10],
            "num_iterations": num_iterations,
            "scale_pos_weight": scale_pos_weight,
            "boost_from_average": True,
            "is_unbalance": True,
            "verbose": -1,
        }
    else:
        params = {
            "max_depth": best_params[0],
            "num_leaves": best_params[1],
            "min_data_in_leaf": best_params[2],
            "feature_pre_filter": False,
            "lambda_l1": best_params[3],
            "lambda_l2": best_params[4],
            "metric": "auc,binary_error",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "binary",
            "learning_rate": learning_rate,
            "feature_fraction": best_params[5],
            "bagging_fraction": best_params[6],
            "min_split_gain": best_params[7],
            "min_child_weight": best_params[8],
            "reg_alpha": best_params[9],
            "reg_lambda": best_params[10],
            "num_iterations": num_iterations,
            "scale_pos_weight": scale_pos_weight,
            "boost_from_average": True,
            "is_unbalance": False,
            "verbose": -1,
        }

    evals_results = {}
    model = lgb.train(
        params,
        d_train,
        valid_sets=[d_train, d_valid],
        valid_names=["train", "valid"],
        evals_result=evals_results,
        num_boost_round=num_boost_round,
        early_stopping_rounds=early_stopping_rounds,
        verbose_eval=None,
        feval=None,
        init_model=mdl_cls_obj,
    )
    test_auc = roc_auc_score(test_lbl_np, model.predict(test_np))
    print("Testing AUC: {}".format(test_auc))
    print("Finish Training")

    model.save_model(out_mdl_file)

    pred_test = model.predict(test_np)
    for i in range(test_np.shape[0]):
        if pred_test[i] >= 0.5:
            pred_test[i] = 1
        else:
            pred_test[i] = 0
    len(pred_test)

    test_acc = accuracy_score(test_lbl_np, pred_test)
    print("Testing Accuracy: {}".format(test_acc))


def apply_lightgbm_binary_classifier(
    model_file,
    in_img_msk,
    img_mask_val,
    img_file_info,
    out_score_img,
    gdalformat,
    out_class_img=None,
    class_thres=5000,
):
    """
    This function applies a trained binary (i.e., two classes) lightgbm model. The function train_lightgbm_binary_classifier
    can be used to train such as model. The output image will contain the softmax score for the class of interest.
    You will need to threshold this image to get a final hard classification. Alternative, a hard class output
    image and threshold can be applied to this image. Note. the softmax score is not a probability.

    :param model_file: a trained lightgbm binary model which can be loaded with lgb.Booster(model_file=model_file).
    :param in_img_msk: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_score_img: output image file with the classification softmax score - this image is scaled by
                       multiplying by 10000.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param out_class_img: Optional output image which will contain the hard classification, defined with a threshold on the
                        probability image.
    :param class_thres: The threshold used to define the hard classification. Default is 5000 (i.e., probability of 0.5).

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")

    def _applyLGBMClassifier(info, inputs, outputs, otherargs):
        outClassVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            outClassVals = outClassVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros(
                (outClassVals.shape[0], otherargs.numClassVars), dtype=numpy.float32
            )
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            predClass = numpy.around(otherargs.classifier.predict(classVars) * 10000)
            outClassVals[ID] = predClass
            outClassVals = numpy.expand_dims(
                outClassVals.reshape(
                    (inputs.imageMask.shape[1], inputs.imageMask.shape[2])
                ),
                axis=0,
            )
        outputs.outimage = outClassVals

    classifier = lgb.Booster(model_file=model_file)

    infiles = applier.FilenameAssociations()
    infiles.imageMask = in_img_msk
    numClassVars = 0
    for imgFile in img_file_info:
        infiles.__dict__[imgFile.name] = imgFile.file_name
        numClassVars = numClassVars + len(imgFile.bands)

    outfiles = applier.FilenameAssociations()
    outfiles.outimage = out_score_img
    otherargs = applier.OtherInputs()
    otherargs.classifier = classifier
    otherargs.mskVal = img_mask_val
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = img_file_info

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(
        _applyLGBMClassifier, infiles, outfiles, otherargs, controls=aControls
    )
    print("Completed")
    rsgislib.imageutils.pop_img_stats(
        out_score_img, use_no_data=True, no_data_val=0, calc_pyramids=True
    )

    if out_class_img is not None:
        rsgislib.imagecalc.image_math(
            out_score_img,
            out_class_img,
            "b1>{}?1:0".format(class_thres),
            gdalformat,
            rsgislib.TYPE_8UINT,
        )
        rsgislib.rastergis.pop_rat_img_stats(
            out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )


def train_lightgbm_multiclass_classifier(
    out_mdl_file,
    cls_info_dict,
    out_info_file=None,
    unbalanced=False,
    n_threads=1,
    early_stopping_rounds=100,
    num_iterations=5000,
    num_boost_round=100,
    learning_rate=0.05,
    mdl_cls_obj=None,
):
    """
    A function which performs a bayesian optimisation of the hyper-parameters for a multiclass lightgbm
    classifier. A dict of class information, as ClassInfoObj objects, is defined with the training data.

    This function requires that lightgbm and skopt modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param out_info_file: An optional output JSON file with information about the classifier which has been created.
    :param unbalanced:
    :param n_threads:
    :param scale_pos_weight:

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")

    from skopt.space import Real, Integer
    from skopt import gp_minimize

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, cls_info_dict[clsname].id
                )
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(cls_info_dict[clsname].test_file_h5, "r")
        sgl_cls_info["test_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["test_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["test_data_lbls"] = numpy.zeros(
            sgl_cls_info["test_n_rows"], dtype=int
        )
        sgl_cls_info["test_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info["test_data"])
        test_lbls_lst.append(sgl_cls_info["test_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)

    print("Finished Reading Data")

    d_train = lgb.Dataset(
        numpy.concatenate(train_data_lst), label=numpy.concatenate(train_lbls_lst)
    )
    d_valid = lgb.Dataset(
        numpy.concatenate(valid_data_lst), label=numpy.concatenate(valid_lbls_lst)
    )

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbl_np = numpy.concatenate(valid_lbls_lst)

    test_np = numpy.concatenate(test_data_lst)
    test_lbl_np = numpy.concatenate(test_lbls_lst)

    space = [
        Integer(3, 10, name="max_depth"),
        Integer(6, 50, name="num_leaves"),
        Integer(10, 50, name="min_data_in_leaf"),
        Real(0, 5, name="lambda_l1"),
        Real(0, 3, name="lambda_l2"),
        Real(0.1, 0.9, name="feature_fraction"),
        Real(0.8, 1.0, name="bagging_fraction"),
        Real(0.001, 0.1, name="min_split_gain"),
        Real(1, 50, name="min_child_weight"),
        Real(1, 1.2, name="reg_alpha"),
        Real(1, 1.4, name="reg_lambda"),
    ]

    def _objective(values):
        if unbalanced:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "multi_logloss",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "multiclass",
                "num_class": n_classes,
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "boost_from_average": True,
                "is_unbalance": True,
            }
        else:
            params = {
                "max_depth": values[0],
                "num_leaves": values[1],
                "min_data_in_leaf": values[2],
                "feature_pre_filter": False,
                "lambda_l1": values[3],
                "lambda_l2": values[4],
                "metric": "multi_logloss",
                "nthread": n_threads,
                "boosting_type": "gbdt",
                "objective": "multiclass",
                "num_class": n_classes,
                "learning_rate": learning_rate,
                "feature_fraction": values[5],
                "bagging_fraction": values[6],
                "min_split_gain": values[7],
                "min_child_weight": values[8],
                "reg_alpha": values[9],
                "reg_lambda": values[10],
                "num_iterations": num_iterations,
                "boost_from_average": True,
                "is_unbalance": False,
            }

        print("\nNext set of params.....", params)

        evals_results = {}
        model_lgb = lgb.train(
            params,
            d_train,
            valid_sets=[d_train, d_valid],
            valid_names=["train", "valid"],
            evals_result=evals_results,
            num_boost_round=num_boost_round,
            early_stopping_rounds=early_stopping_rounds,
            verbose_eval=None,
            feval=None,
            init_model=mdl_cls_obj,
        )

        vld_preds_idxs = numpy.argmax(model_lgb.predict(vaild_np), axis=1)

        acc_score = -accuracy_score(vaild_lbl_np, vld_preds_idxs)
        print(
            "\nAccScore.....",
            -acc_score,
            ".....iter.....",
            model_lgb.current_iteration(),
        )
        gc.collect()
        return acc_score

    res_gp = gp_minimize(
        _objective, space, n_calls=20, random_state=0, n_random_starts=10
    )

    print("Best score={}".format(res_gp.fun))
    best_params = res_gp.x
    print("Best Params:\n{}".format(best_params))

    print("Start Training Find Classifier")
    if unbalanced:
        params = {
            "max_depth": best_params[0],
            "num_leaves": best_params[1],
            "min_data_in_leaf": best_params[2],
            "feature_pre_filter": False,
            "lambda_l1": best_params[3],
            "lambda_l2": best_params[4],
            "metric": "multi_logloss",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "multiclass",
            "num_class": n_classes,
            "learning_rate": learning_rate,
            "feature_fraction": best_params[5],
            "bagging_fraction": best_params[6],
            "min_split_gain": best_params[7],
            "min_child_weight": best_params[8],
            "reg_alpha": best_params[9],
            "reg_lambda": best_params[10],
            "num_iterations": num_iterations,
            "boost_from_average": True,
            "is_unbalance": True,
        }
    else:
        params = {
            "max_depth": best_params[0],
            "num_leaves": best_params[1],
            "min_data_in_leaf": best_params[2],
            "feature_pre_filter": False,
            "lambda_l1": best_params[3],
            "lambda_l2": best_params[4],
            "metric": "multi_logloss",
            "nthread": n_threads,
            "boosting_type": "gbdt",
            "objective": "multiclass",
            "num_class": n_classes,
            "learning_rate": learning_rate,
            "feature_fraction": best_params[5],
            "bagging_fraction": best_params[6],
            "min_split_gain": best_params[7],
            "min_child_weight": best_params[8],
            "reg_alpha": best_params[9],
            "reg_lambda": best_params[10],
            "num_iterations": num_iterations,
            "boost_from_average": True,
            "is_unbalance": False,
        }

    evals_results = {}
    model = lgb.train(
        params,
        d_train,
        valid_sets=[d_train, d_valid],
        valid_names=["train", "valid"],
        evals_result=evals_results,
        num_boost_round=num_boost_round,
        early_stopping_rounds=early_stopping_rounds,
        verbose_eval=None,
        feval=None,
        init_model=mdl_cls_obj,
    )

    vld_preds_idxs = numpy.argmax(model.predict(vaild_np), axis=1)
    valid_acc_scr = accuracy_score(vaild_lbl_np, vld_preds_idxs)
    print("Validate Accuracy: {}".format(valid_acc_scr))
    print("Finish Training")

    model.save_model(out_mdl_file)

    test_preds_idxs = numpy.argmax(model.predict(test_np), axis=1)
    test_acc_scr = accuracy_score(test_lbl_np, test_preds_idxs)
    print("Testing Accuracy: {}".format(test_acc_scr))

    if out_info_file is not None:
        out_info = dict()
        out_info["params"] = params
        out_info["test_accuracy"] = test_acc_scr
        with open(out_info_file, "w") as outfile:
            import json

            json.dump(
                out_info,
                outfile,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                ensure_ascii=False,
            )


def apply_lightgbm_multiclass_classifier(
    cls_train_info,
    model_file,
    in_img_mask,
    img_mask_val,
    img_file_info,
    out_class_img,
    gdalformat,
    class_clr_names=True,
):
    """
    This function applies a trained multiple classes lightgbm model. The function train_lightgbm_multiclass_classifier
    can be used to train such as model. The output image will be a final hard classification using the class with
    the maximum softmax score.

    :param cls_train_info: dict (where the key is the class name) of rsgislib.classification.ClassInfoObj
                           objects which will be used to train the classifier (i.e., train_lightgbm_multiclass_classifier()),
                           provide pixel value id and RGB class values.
    :param model_file: a trained lightgbm multiclass model which can be loaded with lgb.Booster(model_file=model_file).
    :param in_img_mask: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_class_img: Output image which will contain the hard classification defined as the maximum probability.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param class_clr_names: default is True and therefore a colour table will the colours specified in ClassInfoObj
                          and a class_names (from classTrainInfo) column will be added to the output file.

    """
    if not HAVE_LIGHTGBM:
        raise rsgislib.RSGISPyException("Do not have lightgbm module installed.")

    def _applyLGMClassifier(info, inputs, outputs, otherargs):
        outClassVals = numpy.zeros(
            (otherargs.n_classes, inputs.imageMask.shape[1], inputs.imageMask.shape[2]),
            dtype=numpy.uint16,
        )
        outClassIdVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint16)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            n_pxls = inputs.imageMask.shape[1] * inputs.imageMask.shape[2]
            outClassVals = outClassVals.reshape((n_pxls, otherargs.n_classes))
            outClassIdVals = outClassIdVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros(
                (n_pxls, otherargs.numClassVars), dtype=numpy.float32
            )
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            predClassProbs = numpy.around(
                otherargs.classifier.predict(classVars) * 10000
            )
            preds_idxs = numpy.argmax(predClassProbs, axis=1)
            if otherargs.n_classes != predClassProbs.shape[1]:
                raise rsgislib.RSGISPyException(
                    "The number of classes expected and the number provided by the classifier do not match."
                )
            outClassVals[ID] = predClassProbs
            preds_cls_ids = numpy.zeros_like(preds_idxs, dtype=numpy.uint16)
            for cld_id, idx in zip(
                otherargs.cls_id_lut, numpy.arange(0, len(otherargs.cls_id_lut))
            ):
                preds_cls_ids[preds_idxs == idx] = cld_id

            outClassIdVals[ID] = preds_cls_ids
            outClassIdVals = numpy.expand_dims(
                outClassIdVals.reshape(
                    (inputs.imageMask.shape[1], inputs.imageMask.shape[2])
                ),
                axis=0,
            )

        outputs.outclsimage = outClassIdVals

    classifier = lgb.Booster(model_file=model_file)

    infiles = applier.FilenameAssociations()
    infiles.imageMask = in_img_mask
    numClassVars = 0
    for imgFile in img_file_info:
        infiles.__dict__[imgFile.name] = imgFile.file_name
        numClassVars = numClassVars + len(imgFile.bands)

    n_classes = len(cls_train_info)
    cls_id_lut = numpy.zeros(n_classes)
    for clsname in cls_train_info:
        if cls_train_info[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, cls_train_info[clsname].id
                )
            )
        cls_id_lut[cls_train_info[clsname].id] = cls_train_info[clsname].out_id

    outfiles = applier.FilenameAssociations()
    outfiles.outclsimage = out_class_img
    otherargs = applier.OtherInputs()
    otherargs.classifier = classifier
    otherargs.mskVal = img_mask_val
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = img_file_info
    otherargs.n_classes = n_classes
    otherargs.cls_id_lut = cls_id_lut

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Classifier")
    applier.apply(_applyLGMClassifier, infiles, outfiles, otherargs, controls=aControls)
    print("Completed Classification")

    if class_clr_names:
        rsgislib.rastergis.pop_rat_img_stats(
            out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )
        ratDataset = gdal.Open(out_class_img, gdal.GA_Update)
        red = rat.readColumn(ratDataset, "Red")
        green = rat.readColumn(ratDataset, "Green")
        blue = rat.readColumn(ratDataset, "Blue")
        class_names = numpy.empty_like(red, dtype=numpy.dtype("a255"))
        class_names[...] = ""

        for classKey in cls_train_info:
            print("Apply Colour to class '" + classKey + "'")
            red[cls_train_info[classKey].out_id] = cls_train_info[classKey].red
            green[cls_train_info[classKey].out_id] = cls_train_info[classKey].green
            blue[cls_train_info[classKey].out_id] = cls_train_info[classKey].blue
            class_names[cls_train_info[classKey].out_id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "class_names", class_names)
        ratDataset = None



#!/usr/bin/env python
"""
The tools.tilecache module contains functions working with tilecaches.
"""

import math

import rsgislib

TILE_SIZE = 256
EARTH_RADIUS = 6378137.0
ORIGIN_SHIFT = 2.0 * math.pi * EARTH_RADIUS / 2.0
INITIAL_RESOLUTION = 2.0 * math.pi * EARTH_RADIUS / float(TILE_SIZE)


def resolution(zoom):
    """
    Calculate the image resolution for a zoom level.

    :return: float image pixel resolution

    """
    return INITIAL_RESOLUTION / (2**zoom)


def truncate_lng_lat(lon_deg, lat_deg):
    """
    A function which truncates latitude and londitude values
    to be better -180 to 180 and -90 to 90.

    :param lon_deg: float with longitude value.
    :param lat_deg: float with latitude value.
    :return: (lon_deg, lat_deg)

    """
    if lon_deg > 180.0:
        lon_deg = 180.0
    elif lon_deg < -180.0:
        lon_deg = -180.0
    if lat_deg > 90.0:
        lat_deg = 90.0
    elif lat_deg < -90.0:
        lat_deg = -90.0
    return lon_deg, lat_deg


def get_tile_for_point(lon_deg, lat_deg, zoom, tms=True, tile_size=TILE_SIZE):
    """
    :param lon_deg: float with longitude value.
    :param lat_deg: float with latitude value.
    :param zoom: the zoom level for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :param tile_size: is the size, in pixels, of the tiles (default 256).
    :return: (tile_x, tile_y, zoom)

    """
    meter_x = lon_deg * ORIGIN_SHIFT / 180.0
    meter_y = math.log(math.tan((90.0 + lat_deg) * math.pi / 360.0)) / (math.pi / 180.0)
    meter_y = meter_y * ORIGIN_SHIFT / 180.0

    pixel_x = (meter_x + ORIGIN_SHIFT) / resolution(zoom)
    pixel_y = (meter_y - ORIGIN_SHIFT) / resolution(zoom)
    pixel_x = abs(round(pixel_x))
    pixel_y = abs(round(pixel_y))

    tile_x = int(math.ceil(pixel_x / float(tile_size)) - 1)
    tile_y = int(math.ceil(pixel_y / float(tile_size)) - 1)
    if tms:
        tile_y = (2**zoom - 1) - tile_y

    return tile_x, tile_y, zoom


def get_tile_tl_lonlat(tile_x, tile_y, zoom, tms=True):
    """
    Returns the Top-Left coordinate for the tile.

    :param tile_x: The x index for the tile.
    :param tile_y: The y index for the tile (will change if TMS or XYZ indexing).
    :param zoom: the zoom level for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :return: (longitude, latitude)

    """
    if tms:
        tile_y = (2**zoom - 1) - tile_y
    n = 2.0**zoom
    lon_deg = tile_x / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * tile_y / n)))
    lat_deg = math.degrees(lat_rad)
    return (lon_deg, lat_deg)


def get_tile_bbox(tile_x, tile_y, zoom, tms=True):
    """
    Returns the bounding box (BBOX) of a tile

    :param tile_x: The x index for the tile.
    :param tile_y: The y index for the tile (will change if TMS or XYZ indexing).
    :param zoom: the zoom level for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :return: (MinLon, MaxLon, MinLat, MaxLat)

    """
    a = get_tile_tl_lonlat(tile_x, tile_y, zoom, tms)
    b = get_tile_tl_lonlat(tile_x + 1, tile_y + 1, zoom, tms)
    return [a[0], b[0], b[1], a[1]]


def flip_xyz_tms_tiles(tile_x, tile_y, zoom):
    """
    If you put in a XYZ tile index you will have returned
    an TMS index. If you put in a TMS index you will have
    returned XYZ.

    :param tile_x: The x index for the tile.
    :param tile_y: The y index for the tile (will change if TMS or XYZ indexing).
    :param zoom: the zoom level for the tile.
    :return:  (tile_x, tile_y, zoom)

    """
    tile_y = (2**zoom - 1) - tile_y
    return tile_x, tile_y, zoom


def get_web_mercator_xy(lon_deg, lat_deg, truncate=False):
    """
    Convert a WGS84 longitude, latitude coordinate to a web mercator coordinate (meters).

    :param lon_deg: float with longitude value.
    :param lat_deg: float with latitude value.
    :param truncate: Whether to truncate or clip inputs to web mercator limits.
    :return: (meter_x, meter_y)
    """
    if truncate:
        lon_deg, lat_deg = truncate_lng_lat(lon_deg, lat_deg)

    meter_x = lon_deg * ORIGIN_SHIFT / 180.0
    meter_y = math.log(math.tan((90.0 + lat_deg) * math.pi / 360.0)) / (math.pi / 180.0)
    meter_y = meter_y * ORIGIN_SHIFT / 180.0

    return meter_x, meter_y


def get_xy_mercator_lonlat(meter_x, meter_y, truncate=False):
    """
    Convert a web mercator coordinate (meters) to a WGS84 longitude, latitude coordinate.

    :param meter_x: x coordinate in the web mercator coordinate system
    :param meter_y: y coordinate in the web mercator coordinate system
    :param truncate: Whether to truncate or clip inputs to web mercator limits.
    :return: (longitude, latitude)

    """
    lon_deg = (meter_x / ORIGIN_SHIFT) * 180.0
    lat_deg = (meter_y / ORIGIN_SHIFT) * 180.0
    lat_deg = (
        180.0
        / math.pi
        * (2 * math.atan(math.exp(lat_deg * math.pi / 180.0)) - math.pi / 2.0)
    )

    if truncate:
        lon_deg, lat_deg = truncate_lng_lat(lon_deg, lat_deg)

    return lon_deg, lat_deg


def xy_mercator_bounds(tile_x, tile_y, zoom, tms=True):
    """
    Returns the bounding box of a tile in the web mercator coordinate system (meters).

    :param tile_x: The x index for the tile.
    :param tile_y: The y index for the tile (will change if TMS or XYZ indexing).
    :param zoom: the zoom level for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :return: (MinX, MaxX, MinY, MaxY)

    """
    bbox = get_tile_bbox(tile_x, tile_y, zoom, tms)
    x_min, y_max = get_web_mercator_xy(bbox[0], bbox[3])
    x_max, y_min = get_web_mercator_xy(bbox[1], bbox[2])
    return (x_min, x_max, y_min, y_max)


def get_quadkey_for_tile_path(tile_x, tile_y, zoom, tms=True):
    """
    Get the quadkey of the tile

    :param tile_x: The x index for the tile.
    :param tile_y: The y index for the tile (will change if TMS or XYZ indexing).
    :param zoom: the zoom level for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :return: a quadkey for the tile.

    """
    if tms:
        tile_y = (2**zoom - 1) - tile_y
    qk = []
    for z in range(zoom, 0, -1):
        digit = 0
        mask = 1 << (z - 1)
        if tile_x & mask:
            digit += 1
        if tile_y & mask:
            digit += 2
        qk.append(str(digit))
    return "".join(qk)


def get_tile_path_from_quadkey(quadkey, tms=True):
    """
    Return the tile path for an inputted quadkey.

    :param quadkey: the inputted quadkey for the tile.
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :returns: (tile_x, tile_y, zoom)

    """
    if len(quadkey) == 0:
        return 0, 0, 0
    tile_x, tile_y = 0, 0
    for i, digit in enumerate(reversed(quadkey)):
        mask = 1 << i
        if digit == "1":
            tile_x = tile_x | mask
        elif digit == "2":
            tile_y = tile_y | mask
        elif digit == "3":
            tile_x = tile_x | mask
            tile_y = tile_y | mask
        elif digit != "0":
            raise rsgislib.RSGISPyException(
                "Unexpected quadkey digit: {}".format(digit)
            )
    zoom = i + 1
    if tms:
        tile_y = (2**zoom - 1) - tile_y
    return tile_x, tile_y, zoom


def get_tiles(bbox, zooms, tms=True, truncate=False):
    """
    Gets a list of tile indexes for all the zoom levels specified. The tile indexes are returned as a
    dict with the zoom as the key which references a list of tiles.

    :param bbox: A bounding box in WGS84 longitude/latitude values
                 (MinX, MaxX, MinY, MaxY).
    :param zooms: list of zoom levels
    :param tms: if TMS is True then a tile grid in TMS format is returned with
                the grid origin at the bottom-left. If False then an XYZ tile grid
                format is used with the origin in the top-left.
    :param truncate: Whether to truncate or clip inputs to web mercator limits.
    :return: dict[zoom] = [(tile_x, tile_y, zoom),  (tile_x, tile_y, zoom) ...  (tile_x, tile_y, zoom)]

    """
    import math
    from collections.abc import Sequence

    west = bbox[0]
    east = bbox[1]
    south = bbox[2]
    north = bbox[3]
    if truncate:
        west, south = truncate_lng_lat(west, south)
        east, north = truncate_lng_lat(east, north)
    if west > east:
        bbox_west = (-180.0, south, east, north)
        bbox_east = (west, south, 180.0, north)
        bboxes = [bbox_west, bbox_east]
    else:
        bboxes = [(west, south, east, north)]

    tiles = dict()

    for w, s, e, n in bboxes:
        # Clamp bounding values.
        w = max(-180.0, w)
        s = max(-85.051129, s)
        e = min(180.0, e)
        n = min(85.051129, n)

        if not isinstance(zooms, Sequence):
            zooms = [zooms]

        epsilon = 1.0e-9

        for z in zooms:
            if z not in tiles:
                tiles[z] = []
            print(z)
            llx, lly, llz = get_tile_for_point(w, s, z, tms)
            if lly % 1 < epsilon / 10:
                lly = lly - epsilon

            urx, ury, urz = get_tile_for_point(e, n, z, tms)
            if urx % 1 < epsilon / 10:
                urx = urx - epsilon

            # Clamp left x and top y at 0.
            llx = 0 if llx < 0 else llx
            ury = 0 if ury < 0 else ury

            llx, urx, lly, ury = map(lambda x: int(math.floor(x)), [llx, urx, lly, ury])

            min_x_tile = llx if llx < urx else urx

            if llx == urx:
                n_x_tiles = 1
            elif llx < urx:
                n_x_tiles = urx - min(llx + 1, 2**z)
            else:
                n_x_tiles = llx - min(urx + 1, 2**z)

            n_x_tiles = 1 if n_x_tiles == 0 else n_x_tiles
            print("\tx {} + {}".format(min_x_tile, n_x_tiles))

            min_y_tile = ury if ury < lly else lly

            if ury == lly:
                n_y_tiles = 1
            elif ury < lly:
                n_y_tiles = lly - min(ury + 1, 2**z)
            else:
                n_y_tiles = ury - min(lly + 1, 2**z)
            n_y_tiles = 1 if n_y_tiles == 0 else n_y_tiles
            print("\ty {} + {}".format(min_y_tile, n_y_tiles))

            for i in range(min_x_tile, min_x_tile + n_x_tiles):
                for j in range(min_y_tile, min_y_tile + n_y_tiles):
                    tiles[z].append((i, j, z))
    return tiles


def convert_between_tms_xyz(tiles_dir):
    """
    A function which converts between TMS and XYZ tile indexing. The conversion is performed in-situ
    with the files simply renamed. If a tile cache is in TMS format then this function will convert
    it to XYZ, while if it is in XYZ format then it will convert it to TMS (they are easily interchangeable)

    The TMS format uses a grid with the origin at the bottom-left. While the XYZ
    format is uses a grid with the origin in the top-left.

    XYZ grids can be loads into QGIS and other GIS systems as layers.
    TMS grids are outputted by gdal2tiles.py.

    :param tiles_dir: The input directory for the tile cache.

    """
    import os
    import rsgislib.tools.utils
    import tqdm

    c_pwd = os.getcwd()
    in_zoom_dirs = []
    lcl_files = os.listdir(tiles_dir)
    for lcl_file in lcl_files:
        lcl_name = lcl_file
        lcl_file = os.path.join(tiles_dir, lcl_file)
        if os.path.isdir(lcl_file) and rsgislib.tools.utils.is_number(lcl_name):
            in_zoom_dirs.append(os.path.abspath(lcl_file))
    for zoom_dir in tqdm.tqdm(in_zoom_dirs, desc="overall"):
        zoom = int(os.path.basename(zoom_dir))
        x_dirs = os.listdir(zoom_dir)
        for x_dir in tqdm.tqdm(x_dirs, desc="zoom {}".format(zoom), leave=False):
            if rsgislib.tools.utils.is_number(x_dir):
                x = int(x_dir)
                x_dir = os.path.join(zoom_dir, x_dir)
                if os.path.isdir(x_dir) and rsgislib.tools.utils.is_number(x):
                    y_files = os.listdir(x_dir)
                    os.chdir(x_dir)
                    for y_file in tqdm.tqdm(y_files, desc="y", leave=False):
                        y_cur_file = os.path.join(x_dir, y_file)
                        y_in_str = os.path.splitext(y_file)[0]
                        if rsgislib.tools.utils.is_number(y_in_str):
                            y_in = int(y_in_str)
                            img_ext = os.path.splitext(y_file)[1]
                            n_tile_x, n_tile_y, n_zoom = flip_xyz_tms_tiles(
                                x, y_in, zoom
                            )
                            out_img_file = "{}{}".format(n_tile_y, img_ext)
                            os.rename(y_cur_file, out_img_file)
    os.chdir(c_pwd)



import os
import pytest
from shutil import copy2

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
VECTORATTRS_DATA_DIR = os.path.join(DATA_DIR, "vectorattrs")

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True


def test_read_vec_column_IntCol():
    import rsgislib.vectorattrs

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"
    vals = rsgislib.vectorattrs.read_vec_column(vec_file, vec_lyr, "IntCol")
    ref_vals = [1, 2, 3, 4, 5, 6]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_read_vec_column_FloatCol():
    import rsgislib.vectorattrs

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"
    vals = rsgislib.vectorattrs.read_vec_column(vec_file, vec_lyr, "FloatCol")
    ref_vals = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_read_vec_column_StrCol():
    import rsgislib.vectorattrs

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"
    vals = rsgislib.vectorattrs.read_vec_column(vec_file, vec_lyr, "StrCol")
    ref_vals = ["One", "Two", "Three", "Four", "Five", "Six"]
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_read_vec_columns():
    import rsgislib.vectorattrs

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"
    att_columns = ["IntCol", "FloatCol", "StrCol"]
    vals = rsgislib.vectorattrs.read_vec_columns(vec_file, vec_lyr, att_columns)

    ref_int_vals = [1, 2, 3, 4, 5, 6]
    ref_flt_vals = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    ref_str_vals = ["One", "Two", "Three", "Four", "Five", "Six"]
    vals_eq = True
    for i in range(len(vals)):
        if (
            (vals[i]["IntCol"] != ref_int_vals[i])
            and (vals[i]["FloatCol"] != ref_flt_vals[i])
            and (vals[i]["StrCol"] != ref_str_vals[i])
        ):
            vals_eq = False
            break

    assert vals_eq


def test_write_vec_column_Int(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    ref_vals = [1, 2, 3, 4, 5, 6]
    rsgislib.vectorattrs.write_vec_column(
        out_vec_file, out_vec_lyr, "testcol", ogr.OFTInteger, ref_vals
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_write_vec_column_Float(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    ref_vals = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    rsgislib.vectorattrs.write_vec_column(
        out_vec_file, out_vec_lyr, "testcol", ogr.OFTReal, ref_vals
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_write_vec_column_String(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    ref_vals = ["One", "Two", "Three", "Four", "Five", "Six"]
    rsgislib.vectorattrs.write_vec_column(
        out_vec_file, out_vec_lyr, "testcol", ogr.OFTString, ref_vals
    )

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_write_vec_column_to_layer_Int(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    vec_obj_ds, vec_lyr_obj = rsgislib.vectorutils.open_gdal_vec_lyr(
        out_vec_file, out_vec_lyr, readonly=False
    )

    ref_vals = [1, 2, 3, 4, 5, 6]
    rsgislib.vectorattrs.write_vec_column_to_layer(
        vec_lyr_obj, "testcol", ogr.OFTInteger, ref_vals
    )
    vec_obj_ds = None

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_write_vec_column_to_layer_Float(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    vec_obj_ds, vec_lyr_obj = rsgislib.vectorutils.open_gdal_vec_lyr(
        out_vec_file, out_vec_lyr, readonly=False
    )

    ref_vals = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    rsgislib.vectorattrs.write_vec_column_to_layer(
        vec_lyr_obj, "testcol", ogr.OFTReal, ref_vals
    )
    vec_obj_ds = None

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_write_vec_column_to_layer_String(tmp_path):
    import rsgislib.vectorutils
    import rsgislib.vectorattrs
    from osgeo import ogr

    vec_file = os.path.join(VECTORATTRS_DATA_DIR, "sen2_20210527_aber_att_vals.geojson")
    vec_lyr = "sen2_20210527_aber_att_vals"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorutils.create_copy_vector_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        "GEOJSON",
        replace=True,
        in_memory=True,
    )

    vec_obj_ds, vec_lyr_obj = rsgislib.vectorutils.open_gdal_vec_lyr(
        out_vec_file, out_vec_lyr, readonly=False
    )

    ref_vals = ["One", "Two", "Three", "Four", "Five", "Six"]
    rsgislib.vectorattrs.write_vec_column_to_layer(
        vec_lyr_obj, "testcol", ogr.OFTString, ref_vals
    )
    vec_obj_ds = None

    vals = rsgislib.vectorattrs.read_vec_column(out_vec_file, out_vec_lyr, "testcol")
    vals_eq = True
    for val, ref_val in zip(vals, ref_vals):
        if val != ref_val:
            vals_eq = False
            break
    assert vals_eq


def test_pop_bbox_cols(tmp_path):
    import rsgislib.vectorattrs

    vec_in_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_file = os.path.join(tmp_path, "aber_osgb_multi_polys.geojson")
    copy2(vec_in_file, vec_file)

    vec_lyr = "aber_osgb_multi_polys"
    rsgislib.vectorattrs.pop_bbox_cols(vec_file, vec_lyr)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_geom_bbox_cols(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorattrs.add_geom_bbox_cols(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        min_x_col="MinX",
        max_x_col="MaxX",
        min_y_col="MinY",
        max_y_col="MaxY",
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_create_name_col(tmp_path):
    import rsgislib.vectorattrs

    vec_in_file = os.path.join(DATA_DIR, "degree_grid_examples.geojson")
    vec_file = os.path.join(tmp_path, "degree_grid_examples.geojson")
    copy2(vec_in_file, vec_file)
    vec_lyr = "degree_grid_examples"
    rsgislib.vectorattrs.pop_bbox_cols(vec_file, vec_lyr)

    out_vec_file = os.path.join(tmp_path, "degree_grid_named.geojson")
    out_vec_lyr = "degree_grid_named"
    rsgislib.vectorattrs.create_name_col(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GeoJSON",
        out_col="names",
        x_col="xmin",
        y_col="ymax",
        prefix="hello",
        postfix="world",
        coords_lat_lon=True,
        int_coords=True,
        zero_x_pad=3,
        zero_y_pad=2,
        round_n_digts=3,
        non_neg=True,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_unq_numeric_col(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_unq_numeric_col(
        vec_file,
        vec_lyr,
        "names",
        "unq_id",
        out_vec_file,
        out_vec_lyr,
        out_format="GeoJSON",
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_numeric_col_lut(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    val_lut = dict()
    val_lut["helloN63W126world"] = 10
    val_lut["helloN03E134world"] = 20
    val_lut["helloN02W071world"] = 30
    val_lut["helloS51E056world"] = 40
    val_lut["helloS50E049world"] = 50

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_numeric_col_lut(
        vec_file,
        vec_lyr,
        "names",
        val_lut,
        "int_id",
        out_vec_file,
        out_vec_lyr,
        out_format="GeoJSON",
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_numeric_col_float(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_numeric_col(
        vec_file,
        vec_lyr,
        "new_col",
        out_vec_file,
        out_vec_lyr,
        out_val=101,
        out_format="GeoJSON",
        out_col_int=False,
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_numeric_col_float(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_numeric_col(
        vec_file,
        vec_lyr,
        "new_col",
        out_vec_file,
        out_vec_lyr,
        out_val=101,
        out_format="GeoJSON",
        out_col_int=True,
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_string_col(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_string_col(
        vec_file,
        vec_lyr,
        "new_col",
        out_vec_file,
        out_vec_lyr,
        out_val="Hello World",
        out_format="GeoJSON",
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_unq_col_values():
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    rsgislib.vectorattrs.get_unq_col_values(vec_file, vec_lyr, col_name="names")


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_add_fid_col(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"
    rsgislib.vectorattrs.add_fid_col(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GeoJSON",
        out_col="new_fid",
    )
    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_vec_cols_as_array():
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    cols = ["xmin", "xmax", "ymin", "ymax"]
    rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols, lower_limit=None, upper_limit=None
    )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_vec_cols_as_array_lower_limit():
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    cols = ["xmin", "xmax", "ymin", "ymax"]
    rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols, lower_limit=50, upper_limit=None
    )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_vec_cols_as_array_upper_limit():
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    cols = ["xmin", "xmax", "ymin", "ymax"]
    rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols, lower_limit=None, upper_limit=100
    )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_vec_cols_as_array_limits():
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    cols = ["xmin", "xmax", "ymin", "ymax"]
    rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols, lower_limit=50, upper_limit=100
    )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_sort_vec_lyr(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"

    rsgislib.vectorattrs.sort_vec_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        sort_by="names",
        ascending=True,
        out_format="GeoJSON",
    )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_sort_vec_lyr_multi_var(tmp_path):
    import rsgislib.vectorattrs

    vec_file = os.path.join(DATA_DIR, "degree_grid_named.geojson")
    vec_lyr = "degree_grid_named"

    out_vec_file = os.path.join(tmp_path, "out_vec.geojson")
    out_vec_lyr = "out_vec"

    rsgislib.vectorattrs.sort_vec_lyr(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        sort_by=["xmin", "ymin"],
        ascending=[True, True],
        out_format="GeoJSON",
    )



#!/usr/bin/env python
"""
The tools.httptools
"""
from typing import Dict
import json
import os
import requests
import rsgislib
import tqdm


class RSGISPyResponseException(rsgislib.RSGISPyException):
    def __init__(self, value, response=None):
        """
        Init for the RSGISPyResponseException class
        """
        self.value = value
        self.response = response

    def __str__(self):
        """
        Return a string representation of the exception
        """
        return "HTTP status {0} {1}: {2}".format(
            self.response.status_code, self.response.reason, repr(self.value)
        )


def check_http_response(response: requests.Response, url: str) -> bool:
    """
    Check the HTTP response and raise an exception with appropriate error message
    if request was not successful.

    :param response:
    :param url:
    :return:

    """
    try:
        response.raise_for_status()
        success = True
    except (requests.HTTPError, ValueError):
        success = False
        excpt_msg = "Invalid API response."
        try:
            excpt_msg = response.headers["cause-message"]
        except:
            try:
                excpt_msg = response.json()["error"]["message"]["value"]
            except:
                excpt_msg = (
                    "Unknown error ('{0}'), check url in a web browser: '{1}'".format(
                        response.reason, url
                    )
                )
        api_error = RSGISPyResponseException(excpt_msg, response)
        api_error.__cause__ = None
        raise api_error
    return success


def send_http_json_request(url: str, data: Dict = None, api_key: str = None) -> str:
    """
    A function which sends a http request with a json data packet.
    If an error occurs an exception will be raised.

    :param url: The URL for the request to be sent.
    :param data: dictionary of data which can be converted to str
                 using json.dumps.
    :param api_key: if provided then the api-key will be provided
                    via the http header.
    :return: the JSON string returned by the server.

    """
    if data is None:
        json_data = None
    else:
        json_data = json.dumps(data)

    if api_key == None:
        response = requests.post(url, json_data)
    else:
        headers = {"X-Auth-Token": api_key}
        response = requests.post(url, json_data, headers=headers)

    try:
        http_status_code = response.status_code
        if response == None:
            raise rsgislib.RSGISPyException("No output from service")

        output = json.loads(response.text)
        if output["errorCode"] != None:
            raise rsgislib.RSGISPyException(
                "{} - {}".format(output["errorCode"], output["errorMessage"])
            )
        if http_status_code == 404:
            raise rsgislib.RSGISPyException("404 Not Found")
        elif http_status_code == 401:
            raise rsgislib.RSGISPyException("401 Unauthorized")
        elif http_status_code == 400:
            raise rsgislib.RSGISPyException(f"Error Code: {http_status_code}")
    except Exception as e:
        response.close()
        raise rsgislib.RSGISPyException(f"{e}")
    response.close()

    return output["data"]


def download_file_http(
    input_url: str,
    out_file_path: str,
    username: str = None,
    password: str = None,
    no_except: bool = True,
):
    """

    :param input_url:
    :param out_file_path:
    :param username:
    :param password:
    :return:

    """
    session_http = requests.Session()
    if (username is not None) and (password is not None):
        session_http.auth = (username, password)
    user_agent = "rsgislib/{}".format(rsgislib.get_rsgislib_version())
    session_http.headers["User-Agent"] = user_agent

    tmp_dwnld_path = out_file_path + ".incomplete"

    headers = {}

    try:
        with session_http.get(
            input_url, stream=True, auth=session_http.auth, headers=headers
        ) as r:
            check_http_response(r, input_url)
            total = int(r.headers.get("content-length", 0))
            chunk_size = 2**20
            n_chunks = int(total / chunk_size) + 1

            with open(tmp_dwnld_path, "wb") as f:
                for chunk in tqdm.tqdm(
                    r.iter_content(chunk_size=chunk_size), total=n_chunks
                ):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
        os.rename(tmp_dwnld_path, out_file_path)
        print("Download Complete: {}".format(out_file_path))

    except Exception as e:
        if no_except:
            print(e)
        else:
            raise rsgislib.RSGISPyException("{}".format(e))
        return False
    return True



#!/usr/bin/env python
"""
The vector conversion tools for converting between raster and vector
"""

from osgeo import gdal
from osgeo import ogr

import rsgislib

gdal.UseExceptions()


def rasterise_vec_lyr(
    vec_file: str,
    vec_lyr: str,
    input_img: str,
    output_img: str,
    gdalformat: str = "KEA",
    burn_val: int = 1,
    datatype=rsgislib.TYPE_8UINT,
    att_column=None,
    use_vec_extent=False,
    thematic=True,
    no_data_val=0,
):
    """
    A utility to rasterise a vector layer to an image covering the same region and at
    the same resolution as the input image.

    :param vec_file: is a string specifying the input vector file
    :param vec_lyr: is a string specifying the input vector layer name.
    :param input_img: is a string specifying the input image defining the grid, pixel
                      resolution and area for the rasterisation (if None and vecExt is
                      False them assumes output image already exists and just uses it
                      as is burning vector into it)
    :param output_img: is a string specifying the output image for the rasterised
                       vector file
    :param gdalformat: is the output image format (Default: KEA).
    :param burn_val: is the value for the output image pixels if no attribute is
                     provided.
    :param datatype: of the output file, default is rsgislib.TYPE_8UINT
    :param att_column: is a string specifying the attribute to be rasterised, value of
                       None creates a binary mask and \"FID\" creates a temp vector file
                       with a "FID" column and rasterises that column.
    :param use_vec_extent: is a boolean specifying that the output image should be cut
                           to the same extent as the input shapefile (Default is False
                           and therefore output image will be the same as the input).
    :param thematic: is a boolean (default True) specifying that the output image is
                     an thematic dataset so a colour table will be populated.
    :param no_data_val: is a float specifying the no data value associated with a
                        continuous output image.

    .. code:: python

        from rsgislib import vectorutils

        inputVector = 'crowns.shp'
        inputVectorLyr = 'crowns'
        inputImage = 'injune_p142_casi_sub_utm.kea'
        outputImage = 'psu142_crowns.kea'
        vectorutils.rasterise_vec_lyr(inputVector,
                                      inputVectorLyr,
                                      inputImage,
                                      outputImage,
                                      'KEA',
                                      vecAtt='FID')

    """
    import rsgislib.imageutils

    try:
        if use_vec_extent:
            xRes, yRes = rsgislib.imageutils.get_img_res(input_img)
            if yRes < -1:
                yRes = yRes * (-1)
            outRes = xRes
            if xRes > yRes:
                outRes = yRes

            rsgislib.imageutils.create_copy_img_vec_extent_snap_to_grid(
                vec_file, vec_lyr, output_img, outRes, 1, gdalformat, datatype
            )
        elif input_img is None:
            print("Assuming output image is already created so just using.")
        else:
            print("Creating output image using input image")
            rsgislib.imageutils.create_copy_img(
                input_img, output_img, 1, 0, gdalformat, datatype
            )

        print("Running Rasterise now...")
        out_img_ds = gdal.Open(output_img, gdal.GA_Update)
        if out_img_ds is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(output_img))

        vec_ds = gdal.OpenEx(vec_file, gdal.OF_VECTOR)
        if vec_ds is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

        vec_lyr_obj = vec_ds.GetLayerByName(vec_lyr)
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("Could not find layer '{}'".format(vec_lyr))

        # Run the algorithm.
        err = 0
        if att_column is None:
            err = gdal.RasterizeLayer(
                out_img_ds, [1], vec_lyr_obj, burn_values=[burn_val]
            )
        else:
            err = gdal.RasterizeLayer(
                out_img_ds, [1], vec_lyr_obj, options=["ATTRIBUTE=" + att_column]
            )
        if err != 0:
            raise rsgislib.RSGISPyException("Rasterisation Error: " + str(err))

        out_img_ds = None
        vec_ds = None

        if thematic and (gdalformat == "KEA"):
            import rsgislib.rastergis

            rsgislib.rastergis.pop_rat_img_stats(
                clumps_img=output_img,
                add_clr_tab=True,
                calc_pyramids=True,
                ignore_zero=True,
            )
        else:
            rsgislib.imageutils.pop_img_stats(output_img, True, no_data_val, True)
    except Exception as e:
        raise e


def rasterise_vec_lyr_obj(
    vec_lyr_obj: ogr.Layer,
    output_img: str,
    burn_val: int = 1,
    att_column: str = None,
    calc_stats: bool = True,
    thematic: bool = True,
    no_data_val: float = 0,
):
    """
    A utility to rasterise a vector layer to an image covering the same region.

    :param vec_lyr_obj: is a OGR Vector Layer Object
    :param output_img: is a string specifying the output image, this image must already
                     exist and intersect within the input vector layer.
    :param burn_val: is the value for the output image pixels if no attribute is
                     provided.
    :param att_column: is a string specifying the attribute to be rasterised, value
                       of None creates a binary mask and \"FID\" creates a temp
                       vector layer with a "FID" column and rasterises that column.
    :param calc_stats: is a boolean specifying whether image stats and pyramids
                       should be calculated.
    :param thematic: is a boolean (default True) specifying that the output image is
                     an thematic dataset so a colour table will be populated.
    :param no_data_val: is a float specifying the no data value associated with a
                   continuous output image.

    """
    try:
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException(
                "The vec_lyr_obj passed to the function was None."
            )

        print("Running Rasterise now...")
        out_img_ds = gdal.Open(output_img, gdal.GA_Update)
        if out_img_ds is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(output_img))

        # Run the algorithm.
        err = 0
        if att_column is None:
            err = gdal.RasterizeLayer(
                out_img_ds, [1], vec_lyr_obj, burn_values=[burn_val]
            )
        else:
            err = gdal.RasterizeLayer(
                out_img_ds, [1], vec_lyr_obj, options=["ATTRIBUTE=" + att_column]
            )
        if err != 0:
            raise rsgislib.RSGISPyException("Rasterisation Error: {}".format(err))

        out_img_ds = None

        if calc_stats:
            if thematic:
                import rsgislib.rastergis

                rsgislib.rastergis.pop_rat_img_stats(
                    clumps_img=output_img,
                    add_clr_tab=True,
                    calc_pyramids=True,
                    ignore_zero=True,
                )
            else:
                import rsgislib.imageutils

                rsgislib.imageutils.pop_img_stats(output_img, True, no_data_val, True)
    except Exception as e:
        print("Failed rasterising: {}".format(output_img))
        raise e


def copy_vec_to_rat(
    vec_file: str, vec_lyr: str, input_img: str, output_img: str, fid_col: str = "FID"
):
    """
    A utility to create raster copy of a polygon vector layer. The output image is
    a KEA file and the attribute table has the attributes from the vector layer.

    :param vec_file: is a string specifying the input vector file
    :param vec_lyr: is a string specifying the layer within the input vector file
    :param input_img: is a string specifying the input image defining the grid,
                      pixel resolution and area for the rasterisation
    :param output_img: is a string specifying the output KEA image for the
                       rasterised vector layer

    .. code:: python

        from rsgislib import vectorutils

        inputVector = 'crowns.shp'
        inputImage = 'injune_p142_casi_sub_utm.kea'
        outputImage = 'psu142_crowns.kea'

        vectorutils.copy_vec_to_rat(inputVector, 'crowns', inputImage, outputImage)

    """
    import rsgislib
    import rsgislib.rastergis
    import rsgislib.vectorutils

    cols = rsgislib.vectorutils.get_vec_lyr_cols(vec_file, vec_lyr)
    if fid_col not in cols:
        raise rsgislib.RSGISPyException("FID Column not within the input layer")

    rasterise_vec_lyr(
        vec_file,
        vec_lyr,
        input_img,
        output_img,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_32UINT,
        att_column=fid_col,
        use_vec_extent=False,
        thematic=True,
        no_data_val=0,
    )
    rsgislib.rastergis.import_vec_atts(output_img, vec_file, vec_lyr, "pxlval", None)
